//
//  chg.cpp
//  iptv
//
//  Created by Tamás Jursonovics on 2013.07.25..
//  Copyright (c) 2013 Tamás Jursonovics. All rights reserved.
//
//  implements the quality based charging functions inclusive the quality based policy decision and creates the bill and discount for the streaming session.
//


#include <iostream>
#include <fstream>
#include <errno.h>
#include <math.h>
#include <cstring>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>    /* For O_RDWR */
#include <unistd.h>   /* For open(), creat() */


using namespace std;


ifstream meta_stream, rr_stream, si_stream;
int N=0;
float r=0,Q_p,Q_m;

float mse(float psnr)
{
    return pow(255,2)/pow(10,psnr/10);
}

float psnr(float mse)
{
    return 10*log10(pow(255,2)/mse);
}


int main(int argc, const char * argv[])
{

    if (argc!=8)
    {
        cout << "Simple chargin application" << endl<<endl;
        //                              1         2        3            4   5     6    7
        cout<<"Usage: "<<argv[0]<<" <meta_file> <rr_file> <msize_file> <N> <r> <Q_p> <Q_m>"<<endl<<endl;
        cout<<"..."<<endl;        
        return 0;
    }
    
    //1 Parse input args
    sscanf(argv[4],"%d",&N);
    sscanf(argv[5],"%f",&r);
    sscanf(argv[6],"%f",&Q_p);
    sscanf(argv[7],"%f",&Q_m);

    //2. Initialize file
    meta_stream.open(argv[1]);
    if (meta_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[1] << ":" << strerror(errno);
        return -1;
    }
    rr_stream.open(argv[2]);
    if (rr_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[2] << ":" << strerror(errno);
        meta_stream.close();
        return -1;
    }
    si_stream.open(argv[3]);
    if (si_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[3] << ":" << strerror(errno);
        meta_stream.close();
        rr_stream.close();
        return -1;
    }
    
    char meta_line[256];
    char rr_line[256];
    char si_line[256];
    
    //input
    int seq=0, frame=0, k=0, k2=0;
    char hit[256];
    float dk=0, d_PSNRk=0;
    int size=0;
    
    //calculation
    float DSk=0, base=0;
    int i=0;
    int lineno=0;
    float sizesum=0,losssum=0;
    
    //Skip first line
    do rr_stream.getline(rr_line,256); while (rr_line[0]=='#');

    cerr << "#seq" <<"\t"<< "frame" <<"\t"<<"hit" <<"\t"<<"k" <<"\t"<<"dk" <<"\t"<<"d_PSNRk"<<"\t"<<"losssum"<<endl;
    while(rr_stream.good())
    {
        do rr_stream.getline(rr_line,256); while (rr_line[0]=='#');
        do meta_stream.getline(meta_line,256); while (meta_line[0]=='#');
        do si_stream.getline(si_line,256); while (si_line[0]=='#');
        lineno++;
        
        sscanf(rr_line,"%d %d %s",&seq, &frame, &hit);
        sscanf(meta_line,"%d %f %f",&k,&dk,&d_PSNRk);
        sscanf(si_line,"%d %d",&size,&k2);
        
        if(strcmp(hit,"MISS")==0)
        {
            base = dk;
            i=0;
            losssum+=size;
        }
        
        DSk+=pow(r,i)*(1-i/N)*base;
        if(i<N)
        {
            i++;
        }
        else
        {
            base=0;
        }
        
        sizesum+=size;
        
        cerr << seq <<"\t"<< frame <<"\t"<<hit <<"\t"<<k <<"\t"<<dk <<"\t"<<d_PSNRk<<"\t"<<DSk<<"\t"<<losssum<<endl;
    }
    
    float P_linear=0;
    float P_cos=0;
    float P_elliptic=0;
    
    if( psnr(DSk/lineno)<=Q_m )
    {
        P_linear=0;
        P_cos=0;
        P_elliptic=0;
    }
    else if( psnr(DSk/lineno)>=Q_p )
    {
        P_linear=1;
        P_cos=1;
        P_elliptic=1;
    }
    else
    {
        float R=3.14159/2.0-1.0;
        P_linear=(1-R)/(Q_p-Q_m)*psnr(DSk/lineno)+1-(Q_p*(1-R))/(Q_p-Q_m);
        P_cos=(R-1)/2*cos((psnr(DSk/lineno)-Q_m)/(Q_p-Q_m)*3.14159)+R+(1-R)/2;
        P_elliptic=sqrt(1-pow( (psnr(DSk/lineno)-Q_p)/(Q_p-Q_m),2));
    }
    
    std::cout <<"#aDSk\t[dB]\tP_linear\tP_cos\tP_elliptic\tP_size\tfile"<<endl;
    std::cout <<DSk/lineno<<"\t"<<psnr(DSk/lineno)<<"\t"<<P_linear<<"\t"<<P_cos<<"\t"<<P_elliptic<<"\t"<<1-losssum/sizesum<<"\t"<<argv[2]<<endl;

    meta_stream.close();
    rr_stream.close();
    si_stream.close();
    
    return 0;
}


