//
//  chr.cpp
//  iptv
//
//  Created by Tamás Jursonovics on 2013.06.30..
//  Copyright (c) 2013 Tamás Jursonovics. All rights reserved.
//
//  It implements a basic charging gateway functionality. It correlates the Receiver Report files and based on the metadata description it estimates the quality of the streaming session (total distortion: D[κ]) and creates the appropriate Charging Data Records incorporating the quality estimate.
//


#include <iostream>
#include <fstream>
#include "math.h"


using namespace std;


ifstream m_stream;
ifstream l_stream;

float r;
float pi=3.14159,Qp=70,Qm=25,R=pi/2-1;

float u(float x)
{
    if(u<0)
        return 0;
    return 1;
}

float f1(float x){return (-x/pi*2+1)-(-x/pi*2+1)*u(x-pi)  -u(x-pi);}
float f2(float x){return (cos(x)-cos(x)*u(x-pi)  -u(x-pi)  );}
float f3(float x){return (((1-sqrt(1-pow((x/pi-1),2)))*u(-x/pi+1)-1));}

float policy_linear(float x)
{
    return ((R-1)/2*f1((x-Qm)*pi/(Qp-Qm)) + (R+1)/2)*u(x-Qm);
}

float policy_cos(float x)
{
    return ((R-1)/2*f2((x-Qm)*pi/(Qp-Qm)) + (R+1)/2)*u(x-Qm);
}

float policy_elliptic(float x)
{
    return -f3((x-Qm)*pi/(Qp-Qm));
}


int main(int argc, const char * argv[])
{
    if (argc!=4)
    {
        cout << "Simple charging function" << endl<<endl;
        //                              1         2          3
        cout<<"Usage: "<<argv[0]<<" <meta_file> <loss_file> <r>"<<endl<<endl;
        
        return 0;
    }
    
    //0.1 Parse input args
//    strcpy(lst_intf,argv[2]);
//    sscanf(argv[3],"%d",&lst_port);
    sscanf(argv[3],"%f",&r);
    
    
    //2. Initialize file
    m_stream.open(argv[1]);
    if (m_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[1] << ":" << strerror(errno);
        return -1;
    }
    l_stream.open(argv[2]);
    if (l_stream.fail() )
    {
        m_stream.close();
        cerr<<"Error opening the file " << argv[2] << ":" << strerror(errno);
        return -1;
    }
    int number_of_lines = 0;
    char line[256];
    
    while( !l_stream.eof() )
    {
        l_stream.getline(line,256);
        ++number_of_lines;
    }
    l_stream.close();
    l_stream.open(argv[2]);
    if (l_stream.fail() )
    {
        m_stream.close();
        cerr<<"Error opening the file " << argv[2] << ":" << strerror(errno);
        return -1;
    }

//    char m_line[256],l_line[256],status[256];
    string status;
    int frame,seq;
    float dk;
    
    float bprice=100.0/number_of_lines;
    float discount_linear=0;
    float discount_cos=0;
    float discount_elliptic=0;
    
    while(l_stream.good())
    {
        l_stream >> frame;
        l_stream >> seq;
        l_stream >> status;
        if(m_stream.eof())
        {
            m_stream.seekg(0,ios_base::beg);
            m_stream.clear();
        }
        m_stream >> dk;
        
        if( status.compare("MISS")==0 )
        {
            float dk2,d_PSNRk;
            for(int i=0;i<30;i++)
            {
                dk2 = ( pow(r,i) * (1-(i/30)) * dk );
                d_PSNRk=10 * log10(255 * 255 /sqrt(dk2));
                discount_linear -=(bprice * policy_linear(d_PSNRk));
                discount_cos -=(bprice * policy_cos(d_PSNRk));
                discount_elliptic -=(bprice * policy_elliptic(d_PSNRk));
            }
            
        }
        
        
    }
    //Generate CDRs
    cout << "linear   "<<bprice*number_of_lines+discount_linear<<"Ft ("<<bprice*number_of_lines<<"Ft)"<<endl;
    cout << "cos      "<<bprice*number_of_lines+discount_cos<<"Ft ("<<bprice*number_of_lines<<"Ft)"<<endl;
    cout << "elliptic "<<bprice*number_of_lines+discount_elliptic<<"Ft ("<<bprice*number_of_lines<<"Ft)"<<endl;

    m_stream.close();
    l_stream.close();
    
    
    
    return 0;
}


