//
//  cli2.cpp
//  iptv
//
//  Created by Tamás Jursonovics on 2013.06.23..
//  Copyright (c) 2013 Tamás Jursonovics. All rights reserved.
//
//  A simple UDP/RTP receiver which implements the basic client functionalities, streaming server control protocol and an internal jitter buffer. The application issues commands for the streaming server, receives the RTP/UDP stream. It continuously analyses the streaming flow and reports the loss pattern according to the proposed RTSP-RR format to the steaming server.
//


#include <iostream>
#include <fstream>
#include <errno.h>
#include <cstring>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>    /* For O_RDWR */
#include <unistd.h>   /* For open(), creat() */
#include <math.h>   /* For open(), creat() */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <semaphore.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "rtp.h"


char src_addr[256];
int  src_port;
int  fps=0;
int  D_playo=0;

//struct sockaddr_in cli_addr;
int cli_sock=0;

//Timer thread
struct itimerval timer;
struct sigaction sa;
void timer_handler (int signum);

sem_t *s_stop;

#define SIZE 256

typedef struct car_t{
    uint32_t seq;
    int      frame;
    bool     eof;
    
};
car_t root[SIZE];

int seq_off=-1;

using namespace std;


int main (int argc, const char * argv[])
{
    
    cout <<"#Simple UDP/RTP h264 receiver"<<endl;
    if (argc!=5)
    {
        //                              1          2        3      4
        cout<<"Usage: "<<argv[0]<<" <src_addr> <src_port> <fps> <D_playo>"<<endl<<endl;
        return 0;
    }
    strcpy(src_addr,argv[1]);
    src_port=atoi(argv[2]);
    sscanf(argv[3],"%d",&fps);
    D_playo=atoi(argv[4]);
    
    cout << "#src_addr = "<<src_addr<<endl;
    cout << "#src_port = "<<src_port<<endl;
    cout << "#fps      = "<<fps<<endl;
    cout << "#D_playo   = "<<D_playo<<endl;
    
    
    //Create caroussel
    for(int i=0;i<SIZE;i++)
    {
        root[i].seq=0;
        root[i].eof=false;
    }
        
    char name[256];
    
    struct timeval tv;
    gettimeofday(&tv,NULL);
    srand( tv.tv_usec );
    
    sprintf(name,"ser_sem_%d",rand());
    if( (s_stop=sem_open(name, O_CREAT, 0,0)) == SEM_FAILED )
    {
        cout << "Error creating semaphore: " << strerror(errno) <<endl;
        return -1;
    }

    
    //0.2 Configure the timer to expire
    timer.it_value.tv_sec = D_playo/1000;
    timer.it_value.tv_usec =  (D_playo%1000)*1000;
    // ... and after that.
    timer.it_interval.tv_sec = 0;
    timer.it_interval.tv_usec = 1000000/fps;
    
    //0.3. Init timer
    /* Install timer_handler as the signal handler for SIGVTALRM. */
    memset (&sa, 0, sizeof (sa));
    sa.sa_handler = &timer_handler;
    sigaction (SIGALRM, &sa, NULL);

    //initialize control socket
    struct sockaddr_in addr;
    
    if( (cli_sock=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"error creating socket: " << strerror(errno);
        sem_close(s_stop);
        return NULL;
    }
        
    memset(&addr,0,sizeof(addr));
    addr.sin_family=AF_INET;
    addr.sin_port=htons((unsigned long)src_port);
    addr.sin_addr.s_addr=inet_addr(src_addr); /* N.B.: differs from sender */
    
    if ( addr.sin_addr.s_addr==INADDR_NONE )
        addr.sin_addr.s_addr = INADDR_ANY;
    
    
    
    int n=0;
    if( (n=sendto(cli_sock,"hello",5,NULL,(const struct sockaddr *)&addr,sizeof(struct sockaddr_in))) < 0 )
    {
        if( (errno != EINTR)  )
            cerr<<"error sending datagram: " << strerror(errno) << endl;
        cerr.flush();
        sem_close(s_stop);
    }
    
    cout << "# seq  |  frame"<<endl;

    char* buff=(char*)malloc(1024*1025*sizeof(char));
    int slen;
    while(true)
    {
        if( (n=recvfrom(cli_sock, buff,1024*1025,0,(struct sockaddr*)&addr,(socklen_t *)&slen)) != -1 )
        {
            //Size?
            if( n<sizeof(rtp_hdr_t) )
            {
                cerr << "too small packet"<<endl;
                continue;
            }

            //Version?
            if( ((rtp_hdr_t*)buff)->version != 2 )
            {
                cerr << "wrong RTP version"<<endl;
                continue;
            }
            
            //End?
            if( strncmp(&buff[16],"bye",3)==0 )
            {
                root[ntohs(((rtp_hdr_t*)buff)->seq) % SIZE].eof=true;
                break;
            }


            if( seq_off == -1)
            {
                seq_off = ntohs(((rtp_hdr_t*)buff)->seq);

                //5. Start a virtual timer. It counts down whenever this process is executing.
                if (setitimer (ITIMER_REAL, &timer, NULL) != 0 )
                {
                    cout << "error stting timer: " << strerror(errno);
                    return -1;
                }
            }

            //Store
            root[ntohs(((rtp_hdr_t*)buff)->seq) % SIZE].seq=ntohs(((rtp_hdr_t*)buff)->seq);
            root[ntohs(((rtp_hdr_t*)buff)->seq) % SIZE].frame=atoi((char*)&buff[16+9]);
        }
    }
    
    close(cli_sock);

    while((sem_trywait(s_stop)==-1) && (errno==EAGAIN));
    
    free(buff);
    sem_close(s_stop);
    return 0;
}



void timer_handler (int signum)
{
    static int read=0;
    static uint32_t lastseq=-1;
    static bool first=true;
    
    if( first )
    {
        //Hit!!!
        printf("%05d %10d HIT\n",root[read].frame,root[read].seq);
        lastseq=root[read].seq;
        read++;
        first=false;
        return;
    }
    
    if( root[read].eof )
    {
        sem_post(s_stop);
        return;
    }

    if( root[read].seq>lastseq )
    {
        //Hit!!!
        printf("%05d %10d HIT\n",root[read].frame,root[read].seq);
        lastseq=root[read].seq;
    }
    else
    {
        //Miss!!!
        printf("00000 %10d MISS\n",lastseq+1);
        lastseq+=1;
    }
    cout.flush();
    read++;
    read %= SIZE;
}
