//
//  iptv.h
//  iptv
//
//  Created by Tamas Jursonovics on 2012.06.30..
//  Copyright (c) 2012 Tamas Jursonovics. All rights reserved.
//

#ifndef iptv_iptv_h
#define iptv_iptv_h

#include <iostream>
#include <fstream>

#define HEAD_START_CODE    0x010000  //little endian (reverse order for Intel)
#define PICTURE_START_CODE 0x00
#define SLICE_MIN_START    0x01
#define SLICE_MAX_START    0xAF
#define USER_START_CODE    0xB2
#define SEQ_START_CODE     0xB3
#define EXT_START_CODE     0xB5
#define SEQ_END_CODE       0xB7
#define GOP_START_CODE     0xB8

#endif

unsigned long getnal(std::ifstream & mpeg_input, unsigned char* nal)
{
    static unsigned long head_value=0xffffffff;
    unsigned char * a;
    a=(unsigned char *)&head_value;
    
    
    unsigned int pos=-1;
    
    while (mpeg_input)
    {   
        //NAL start?
        if ((head_value&0x00FFFFFF)==HEAD_START_CODE)
        {
            if( pos == -1 )
            {
                //First nal, save nal byte
                pos=0;
            }
            else
            {
                //Second nal, exit
                return pos-1;
            }
        }
        
        if( pos != -1)
        {
            //record nal
            nal[pos]=a[3];
            pos++;
        }   
        
        //Read byte
        head_value=head_value>>8;
        a[3]=mpeg_input.get();
        
    }
    
    // End of file reached
    //std::cout<<" End of broked file at "<<pos<<std::endl;
    return pos;
}

bool getRawBit(unsigned int *bitcounter, unsigned char* RBSP_data)
{
    unsigned int returnValue=0;
    unsigned int count=0;
    unsigned char RBSP_current_byte=0;
    while(count<=(*bitcounter))
    {
        //8-byte fast forward 
        if (((*bitcounter)-count)>7)
        {
            count+=8;
            RBSP_current_byte++;
            RBSP_data=&RBSP_data[RBSP_current_byte];
            continue;
        }
        
        //Classic bit by bit loading
        returnValue=((RBSP_data[RBSP_current_byte]>>(7-(*bitcounter)))&1);
        count++;
    }
    
    (*bitcounter)++;   
    return returnValue;
}
//Decoder functions

unsigned int expGolomb_UD(unsigned int *bitcounter, unsigned char* RBSP_data)
{
    unsigned int M = 0;
    unsigned int info = 0;
    bool bit = 0;
    unsigned int val = 0;
    
    do
    {
        bit = getRawBit(bitcounter,RBSP_data);
        if ( !bit )
            M++;
    }
    while( !bit && M<64 );//terminate if the number is too big!
    
    //now get the M info bits       
    for ( unsigned int i=0 ; i<M ; ++i)
    {
        bit = getRawBit(bitcounter,RBSP_data);
        if ( bit==1 ) 
        {
            //1 jött
            info |= (1<<i);
        }
        else
        {
            //0 jött
            info=info<<1;
        }
    }// i    
    val = (1<<M) + info;
    val -=1;
    
    return val;
}



