//
//  ser2.cpp
//  iptv
//
//  Created by Tamás Jursonovics on 2013.06.23..
//  Copyright (c) 2013 Tamás Jursonovics. All rights reserved.
//
//  A simple H.264 packetizer, generates UDP/RTP multicast traffic based on the RFC 6184 and implements a simple control protocol.
//


#include <iostream>
#include <fstream>
#include <errno.h>
#include <cstring>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>    /* For O_RDWR */
#include <unistd.h>   /* For open(), creat() */

#include <semaphore.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <pthread.h>

#include "rtp.h"


using namespace std;


char lst_intf[256];
int  lst_port=0;
int  fps=0;

struct sockaddr_in cli_addr;
int ser_sock=0;

sem_t *s_takt,*s_start,*s_stop;

//Timer thread
struct itimerval timer;
struct sigaction sa;
void timer_handler (int signum);

//Controller thread
void *th_controller(void *arg);
pthread_mutex_t m_controller;
int change=0;

char *rtp=NULL;

void lipsumgen(char* dst,int length,int frame);

ifstream in_stream;


int main (int argc, const char * argv[])
{
    if (argc!=5)
    {
        cout << "Simple UDP/RTP cbr packetizer" << endl<<endl;
        //                              1         2          3         4
        cout<<"Usage: "<<argv[0]<<" <h264_file> <lst_intf> <lst_port> <fps>"<<endl<<endl;
        cout<<"Dummy UDP/RTP packetizes. Listen for commands on <lst_port> and streams back to the port from which the hello command arrived with <fps> rate."<<endl;
        cout << "Commands: \"hello\" : start"<<endl;
        cout << "          \"bye\"   : exit"<<endl;
//        cout << "          \"up\"    : switch to next file"<<endl;
//        cout << "          \"down\"  : switch to previous file"<<endl;
        cout << "(you may use nc utility for a human intercafe)"<<endl;
        
        return 0;
    }
    
    //0.1 Parse input args
    strcpy(lst_intf,argv[2]);
    sscanf(argv[3],"%d",&lst_port);
    sscanf(argv[4],"%d",&fps);

    //0.2 Configure the timer to expire
    timer.it_value.tv_sec = 0;
    timer.it_value.tv_usec =  1000000/fps;
    // ... and after that.
    timer.it_interval.tv_sec = 0;
    timer.it_interval.tv_usec = 1000000/fps;
    
    //0.3. Init timer
    /* Install timer_handler as the signal handler for SIGVTALRM. */
    memset (&sa, 0, sizeof (sa));
    sa.sa_handler = &timer_handler;
    sigaction (SIGALRM, &sa, NULL);
    
    //0.4 Lipsum
    rtp=(char*)malloc(1025*1024*sizeof(char));

    //Init RTP header
    ((rtp_hdr_t*)rtp)->version=2;
    ((rtp_hdr_t*)rtp)->p=0;
    ((rtp_hdr_t*)rtp)->x=0;
    ((rtp_hdr_t*)rtp)->cc=1;
    ((rtp_hdr_t*)rtp)->m=0;
    ((rtp_hdr_t*)rtp)->pt=96;
    ((rtp_hdr_t*)rtp)->seq=htons(0);//(srand(time(NULL)));
    ((rtp_hdr_t*)rtp)->ts=0;
    ((rtp_hdr_t*)rtp)->ssrc=htonl(0x12345678);
    ((rtp_hdr_t*)rtp)->csrc[0]=htonl(0x87654321);


    //1. Create semaphores for takt, start, stop
    char name[256];

    struct timeval tv;
    gettimeofday(&tv,NULL);
    srand( tv.tv_usec );
    
    sprintf(name,"ser_sem_%d_%d",rand(),lst_port);
    if( (s_takt=sem_open(name, O_CREAT, 0,0)) == SEM_FAILED )
    {
        cout << "Error creating semaphore: " << strerror(errno) <<endl;
        free(rtp);
        return -1;
    }
    sprintf(name,"ser_sem_%d_%d",rand(),lst_port);
    if( (s_start=sem_open(name, O_CREAT, 0,0)) == SEM_FAILED )
    {
        cout << "Error creating semaphore: " << strerror(errno) <<endl;
        free(rtp);
        sem_close(s_takt);
        return -1;
    }
    sprintf(name,"ser_sem_%d_%d",rand(),lst_port);
    if( (s_stop=sem_open(name, O_CREAT, 0,0)) == SEM_FAILED )
    {
        cout << "Error creating semaphore: " << strerror(errno) <<endl;
        free(rtp);
        sem_close(s_takt);
        sem_close(s_start);
        return -1;
    }
    
    //2. Initialize file
    in_stream.open(argv[1]);
    if (in_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[1] << ":" << strerror(errno);
        free(rtp);
        sem_close(s_takt);sem_close(s_start);sem_close(s_stop);

        return -1;
    }
    
    //3. Create controller thread
    pthread_mutex_init(&m_controller,NULL);
    
    pthread_t thcontroller;
    if( pthread_create(&thcontroller, NULL, th_controller, (void *)atoi(argv[3])) )
    {
        cout<<"Error creatig thread: " << strerror(errno);
        free(rtp);
        sem_close(s_takt);sem_close(s_start);sem_close(s_stop);
        in_stream.close();
        pthread_mutex_destroy(&m_controller);
        return -1;
    }

    //4. Wait for start
    sem_wait(s_start);
    
    //5. Start a virtual timer. It counts down whenever this process is executing.
    if (setitimer (ITIMER_REAL, &timer, NULL) != 0 )
    {
        cout << "error stting timer: " << strerror(errno);
        free(rtp);
        sem_close(s_takt);sem_close(s_start);sem_close(s_stop);
        in_stream.close();
        pthread_mutex_destroy(&m_controller);
        return -1;
    }
        
    //6. Start cycle
    int n=0;
    char line[256];
    int frame=0;
//    bool stop=false;
    while(true)
    {
        //7. Wait for takt
        if(sem_wait(s_takt)!=0)
            continue;

        in_stream.getline(line,256);
        if( in_stream.eof() )
        {
            in_stream.close();
            strncpy(&rtp[16],"bye",3);
            sendto(ser_sock,rtp,16+3,NULL,(const struct sockaddr *)&cli_addr,sizeof(struct sockaddr_in));

            sem_post(s_stop);
            break;
        }
        
        lipsumgen(rtp+16,atoi(line),frame);
        
        if( (n=sendto(ser_sock,rtp,16+atoi(line),NULL,(const struct sockaddr *)&cli_addr,sizeof(struct sockaddr_in))) < 0 )
        {
            if( (errno != EINTR) && (errno != EAGAIN) && (errno != EWOULDBLOCK) )
                cerr<<"error sending datagram: " << strerror(errno) << endl;
                cerr.flush();
        }
        
        //step rtp
        ((rtp_hdr_t*)rtp)->seq=htons(ntohs(((rtp_hdr_t*)rtp)->seq)+1);
        ((rtp_hdr_t*)rtp)->ts=htonl(ntohl(((rtp_hdr_t*)rtp)->ts)+90000/fps);
        frame++;
    }
    
    //8. Close
    pthread_mutex_destroy(&m_controller);
    sem_close(s_takt);sem_close(s_start);sem_close(s_stop);
    free(rtp);
}


void timer_handler (int signum)
{
    //Increment semaphore
    sem_post(s_takt);
}


void *th_controller(void *arg)
{
    //initialize control socket
    struct sockaddr_in addr;
    int addrlen=sizeof(addr);
    
    if( (ser_sock=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"error creating socket: " << strerror(errno);
        sem_post(s_stop);
        return NULL;
    }
    
    int flags=0;
    flags = fcntl(ser_sock,F_GETFL);
//    flags |= O_NONBLOCK;
    
    int yes = 1;
    if ( (setsockopt (ser_sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof( yes )) < 0))
    {
        cout<<"error Reusing ADDR: " << strerror(errno);
        sem_post(s_stop);
        return NULL;
    }

    memset(&addr,0,sizeof(addr));
    addr.sin_family=AF_INET;
    addr.sin_port=htons((unsigned long)arg);
    addr.sin_addr.s_addr=inet_addr(lst_intf); /* N.B.: differs from sender */
    
    if ( addr.sin_addr.s_addr==INADDR_NONE )
        addr.sin_addr.s_addr = INADDR_ANY;
    
    if( bind(ser_sock,(struct sockaddr *) &addr,sizeof(addr)) != 0)
    {
        cout << "error binding socket: " << strerror(errno);
        close(ser_sock);
        sem_post(s_stop);
        return NULL;
    }
    
    fcntl(ser_sock, F_SETFL, flags);

    
    char buff[20000];
    int n;
    
    while((sem_trywait(s_stop)==-1) && (errno==EAGAIN))
    {
        if( (n=recvfrom(ser_sock,buff,20000,0,(struct sockaddr *) &cli_addr,(socklen_t *)&addrlen)) < 0 )
        {
            if( (errno != EINTR) && (errno != EAGAIN) && (errno != EWOULDBLOCK) )
                cerr << "error retrieving datagram: " << strerror(errno) << endl;
                cerr.flush();
        }
        else
        {
            buff[n]='\0';
            
            //Parse packet
            if( strncmp("hello",buff,5) == 0 )
                sem_post(s_start);
        }
    }
    
    close(ser_sock);
    return NULL;
}


void lipsumgen(char* dst,int length, int frame)
{
    if(length<=0)
        return;
    
    if(length > 1024*1024)
        return;
    
    int i;
    for(i=0;i<length-1;i++)
        dst[i]=49 + rand() % (126 - 48);
    
    sprintf(&dst[6],"-->%05d<--",frame);
    dst[i]='\n';
}
