//
//  cli.cpp
//  iptv
//
//  Created by Tamás Jursonovics on 2012.05.13..
//  Copyright (c) 2012 Tamás Jursonovics. All rights reserved.
//
//  Implements a simple multicast receiver and control functions of the Multicast Content Delivery Function and RET Server.
//


#include <iostream>
#include <iomanip>
#include <fstream>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include "rtp.h"
#include "nal.h"
#include <semaphore.h>
#include <math.h>

#include <signal.h>
#include <unistd.h>

#include <sys/time.h>
#include <string.h>
#include <errno.h>

#define CLOCKID CLOCK_REALTIME
#define SIG SIGRTMIN

#define GREY   ""//"\033[37m"      /* Bold Black */
#define RESET  ""// "\033[0m"
#define DOWNUP   "\033[4S\033[4A"


#define GOOD 0
#define LOST 1
#define RET  2
#define RETSKIP 3


char src_addr[256];
int  src_port;
char src_intf[256];
int  src_sock;

char ret_addr[256];
int  ret_port;
int  ret_sock;
struct sockaddr_in ret_sockaddr;

char cnt_addr[256];
int  cnt_port;
int  sock_cnt;
struct sockaddr_in cnt_sockaddr;


#define WINDOW (90*20) // 20sec !!!ZÁRÓJEL!!!!  packet
struct markovi_t{
    double B_av;      //Bpsec
    double B_ret;     //Bpsec
    double D_playout; //sec
    double RTT;       //sec
    double mtu;       //sec
    double AVG_psize; //byte
    double D_pkg;     //sec
    double D_ret;     //sec

    double k_Rgeneralmax;
    double k_Rburstmax;
}m_in;

struct markovo_t{

    //Steady state probabilities
    double P_Lsteady;
    double rP_Lsteady;
    int    rN_L;
    double P_Rsteady;   
    double rP_Rsteady;
    
    //Steady state burst lengths
    double P_Lburst[10]; //Maxburst
    double rP_Lburst[10]; //Maxburst
    double P_Rburst[10]; //Maxburst
    double rP_Rburst[10]; //Maxburst
    
    //Rate adapt
    double p_20;
    double p_21;
    

    bool   skip;
    
    double req_B_ret;  //bps
    
    int    n_Rinterburstmin;

};


//Timer thread for packet timing
static void
handler(int sig, siginfo_t *si, void *uc)
{
    /* Note: calling printf() from a signal handler is not
     strictly correct, since printf() is not async-signal-safe;
     see signal(7) */
    
    printf("Caught signal %d\n", sig);
    //print_siginfo(si);
    signal(sig, SIG_IGN);
}

void *th_timer(void *arg);
void timer_handler (int signum);
sem_t *s_stop;

void markov(unsigned state, markovi_t m_in, markovo_t* m_out, bool retskip);


//Playout buffer
#define MAXFRAMESIZE 16 //byte 
struct playout_t{
    u_int8 buff[MAXFRAMESIZE]; 
    int size;
    bool retreq;
    bool retskip;
    rtp_hdr_t *prtp;
    nal_hdr_t *pnal;
};
playout_t *caroussel=NULL;

pthread_mutex_t m_playo;

//MTU
#define MTU 1400
int PLAYOSIZE=100;

pthread_mutex_t m_cread;

bool stop=false;

int fps=0;

using namespace std;


int main (int argc, const char * argv[])
{

    cout <<"#Simple adaptiv UDP/RTP h264 client"<<endl;
    if (argc!=10)
    {
        //                              1          2          3          4          5          6
        //     7           8              9  
        cout<<"Usage: "<<argv[0]<<" <mcs_addr> <mcs_port> <mcs_intf> <cnt_addr> <cnt_port> <ret_addr> <ret_port> <h264_video_file> <fps>"<<endl<<endl;
        cout<<"Listens to <mcs_addr>:<mcs_port> UDP/RTP multicast stream on interface <mcs_intf>, sends adaptation commands to <cnt_addr>:<cnt_port>. Stream is saved to <h264_video_file>. Expects <fps>."<<endl;
        cout<<"Error correction on <ret_addr>:<ret_port>."<<endl;
        return 0;
    }
    strcpy(src_addr,argv[1]);
    src_port=atoi(argv[2]);
    strcpy(src_intf,argv[3]);

    strcpy(ret_addr,argv[6]);
    ret_port=atoi(argv[7]);

    strcpy(cnt_addr,argv[4]);
    cnt_port=atoi(argv[5]);
    
    sscanf(argv[9],"%d",&fps);

    cout << "#mcs_addr = "<<src_addr<<endl;
    cout << "#mcs_port = "<<src_port<<endl;
    cout << "#mcs_intf = "<<src_intf<<endl;
    cout << "#cnt_addr = "<<cnt_addr<<endl;
    cout << "#cnt_port = "<<cnt_port<<endl;
    cout << "#ret_addr = "<<ret_addr<<endl;
    cout << "#ret_port = "<<ret_port<<endl;
    cout << "#file     = "<<argv[8]<<endl;
    cout << "#window   = "<<WINDOW<<endl;
    cout << "#fps      = "<<fps<<endl;
    
    //0. Markov in parameters
    m_in.B_av       =(1000*1000);        //bpsec
    m_in.B_ret      =(120*1000);   //bpsec
    m_in.D_playout  =1.0;                //sec
    m_in.RTT        =0.006;              //sec
    m_in.mtu        =MTU*8;              //bit
    m_in.AVG_psize  =m_in.B_av/fps/3/2;       //bit
    m_in.D_pkg      =(m_in.AVG_psize/m_in.B_av);   //sec
    m_in.D_ret      =(m_in.AVG_psize/m_in.B_ret);  //sec

    PLAYOSIZE=m_in.B_av/(1400*8)*2;
    

    
    cout << "#"<<endl;
    cout << "#Inputs:"<<endl;
    cout << "#B_av     = "<<m_in.B_av/1000<<" kbps"<<endl;
    cout << "#B_ret    = "<<m_in.B_ret/1000<<" kbps"<<endl;
    cout << "#D_playout= "<<m_in.D_playout<<" sec"<<endl;
    cout << "#RTT      = "<<m_in.RTT<<" sec"<<endl;
    cout << "#MTU      = "<<m_in.mtu/8<<" byte"<<endl;
    cout << "#"<<endl;
    cout << "#Outputs:"<<endl;
    cout << "#AVG_psize= "<<m_in.AVG_psize/8<<" byte"<<endl;
    cout << "#D_pkg    = "<<m_in.D_pkg<<" sec"<<endl;
    cout << "#D_ret    = "<<m_in.D_ret<<" sec"<<endl;
    cout << "#frameb   = "<<PLAYOSIZE<<endl;
    cout << "#"<<endl;


    //1. Initiallize playout buffer
    caroussel=(playout_t*)malloc(sizeof(playout_t)*PLAYOSIZE);

    
    for(int i=0;i<PLAYOSIZE;i++)
    {
        caroussel[i].size=0;
        caroussel[i].retreq=false;
        caroussel[i].retskip=false;
        caroussel[i].prtp=(rtp_hdr_t*)caroussel[i].buff;
        caroussel[i].pnal=(nal_hdr_t*)&(caroussel[i].buff[sizeof(rtp_hdr_t)]);
        //Fu header
    }

 
    //2. Create semaphore for exit
    char name[256];
    srand( time(NULL) );
    sprintf(name,"cli_semt_%d",rand());
    if( (s_stop=sem_open(name,O_CREAT|O_EXCL,644,0)) == SEM_FAILED )
    {
        cout << "Error creating semaphore: " << strerror(errno) <<endl;
        return -1;
    }
    
    //3. Configure timer
    struct sigaction sa;
    struct itimerval timer;
    
    // Install timer_handler as the signal handler for SIGVTALRM.
    memset (&sa, 0, sizeof (sa));
    sa.sa_handler = &timer_handler;
    sigaction (SIGALRM, &sa, NULL);
    
    // Configure the timer to expire
    timer.it_value.tv_sec = 1/fps;
    timer.it_value.tv_usec =    (1000000/fps) % 1000000;
    /* ... and after that. */
    timer.it_interval.tv_sec = 1/fps;
    timer.it_interval.tv_usec = (1000000/fps) % 1000000;


    //4. Initialize control socket
    struct hostent *hp;
    
    if ((sock_cnt= socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"error creating socket: " << strerror(errno);
        sem_close(s_stop);
        return -1;
    }
    
    cnt_sockaddr.sin_family = AF_INET;
    if ((hp = gethostbyname(argv[4])) == 0)
    {
        cout<<"Unknown host: " << argv[4] << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        return -1;
    }
    
    bcopy((char *)hp->h_addr, 
          (char *)&cnt_sockaddr.sin_addr,
          hp->h_length);
    cnt_sockaddr.sin_port = htons(atoi(argv[5]));
    
    
    //4b. Initialize retransmission socket
#define TCP
#ifndef TCP
    if ((ret_sock= socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"error creating socket: " << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        return -1;
    }
#endif
    
    ret_sockaddr.sin_family = AF_INET;
    if ((hp = gethostbyname(ret_addr)) == 0)
    {
        cout<<"Unknown host: " << ret_addr << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        return -1;
    }
    
    bcopy((char *)hp->h_addr, 
          (char *)&ret_sockaddr.sin_addr,
          hp->h_length);
    ret_sockaddr.sin_port = htons(ret_port);


    
    //5. Initialize receive socket
    struct sockaddr_in src_sockaddr;
    struct ip_mreq group;
    int n;
    
    // Create a datagram socket on which to receive.
    if( (src_sock=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"error creating socket: " << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        return -1;
    }
    
    int flags = fcntl(src_sock, F_GETFL);
    flags |= O_NONBLOCK;
    fcntl(src_sock, F_SETFL, flags);
    
    // Enable SO_REUSEADDR to allow multiple instances of this
    // application to receive copies of the multicast datagrams.
    int reuse = 1;
    if(setsockopt(src_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof(reuse)) < 0)
    {
        cout<<"error reuse socket: " << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        close(src_sock);
        return -1;
    }
    
    /* Bind to the proper port number with the IP address */
    /* specified as INADDR_ANY. */
    memset((char *) &src_sockaddr, 0, sizeof(src_sockaddr));
    src_sockaddr.sin_family = AF_INET;
    src_sockaddr.sin_addr.s_addr = inet_addr(src_addr);
    src_sockaddr.sin_port = htons(src_port);
    if(bind(src_sock, (struct sockaddr*)&src_sockaddr, sizeof(src_sockaddr)))
    {
        cout<<"error bind socket: " << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        close(src_sock);
        return -1;
    } 
    
    /* Join the multicast group 226.1.1.1 on the local 203.106.93.94 */
    /* interface. Note that this IP_ADD_MEMBERSHIP option must be */
    /* called for each local interface over which the multicast */
    /* datagrams are to be received. */
    group.imr_multiaddr.s_addr = inet_addr(src_addr);
    group.imr_interface.s_addr = inet_addr(src_intf);
    if(setsockopt(src_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&group, sizeof(group)) < 0)
    {
        cout<<"error mcast join socket: " << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        close(src_sock);
        return -1;
    } 
    
    
    //6. Ready, ask for stream!
    n=sendto(sock_cnt,"hello\n",6,0,(const struct sockaddr *)&cnt_sockaddr,sizeof(struct sockaddr_in));
    if (n != 6)
    {
        cout<<"error sending hello: " << strerror(errno) << endl;   
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        close(src_sock);
        return -1;
    }
    cerr << "hello...";
    

    //7. Start a virtual timer. It counts down whenever this process is executing. 
    if (setitimer (ITIMER_REAL, &timer, NULL) != 0 )
    {
        cout << "error setitimer timer: " << strerror(errno);
        sem_close(s_stop);
        close(sock_cnt);
        close(ret_sock);
        close(src_sock);
        return -1;
    }

    //8. Working...
    int ret=0;
    do{
        ret = sem_wait(s_stop);
    }
    while( (ret == -1) && (errno == EINTR) );

    
    //Bye
    n=sendto(sock_cnt,"bye\n",4,0,(const struct sockaddr *)&cnt_sockaddr,sizeof(struct sockaddr_in));
    if (n != 4)
    {
        cerr<<"error sending bye: " << strerror(errno) << endl;   
    }
    cerr << "bye...";

    
    sem_close(s_stop);
    close(sock_cnt);
    close(ret_sock);
    close(src_sock);
    
    return 0;
}


void timer_handler (int signum)
{
    //Initialize
    static char buff[MTU];
    static int posin=-1,posinmax=-1;
    
    static int seqin=-1 ,seqinmax=-1, seqoff=-1;;
    

    static int countdown_out=fps;
    
    
    static int seqout=-1;
    
    static int tsout=-1;

    static struct sockaddr_in src;
    static int slen=sizeof(src);
    
    static int n;
    
    static markovo_t m_out;
    
    static int AVG_psize_num=0;
    static double AVG_psize=0;
    
    static int stat_pps=0; //received (0+2) packet per sec
    static int stat_Bps=0; //received (0+2) byte per sec

    static int stat_lpps=0; //lost (1) packet per sec
    static int stat_time=0; //lost (1) packet per sec


    //1. Receive packet(s) first (may retransmitted packets arrive)
    while( (n=recvfrom(src_sock, buff, sizeof(buff),0,(struct sockaddr*)&src,(socklen_t *)&slen)) != -1 )
    {
        //Size?
        if( n<sizeof(rtp_hdr_t) )
        {
            cout << "too small packet"<<endl;
            continue;
        }
            
        //Version?
        if( ((rtp_hdr_t*)buff)->version != 2 )
        {
//            cout << "wrong RTP version"<<endl;
            continue;
        }
            
        //Adjust seq offset
        if( seqoff == -1 )
        {
            seqoff = ntohs(((rtp_hdr_t*)buff)->seq);
            continue;
        }
        
        //Put into the caroussel (Based on RTP TIME)
        #define SEQ2POS(S) ((S - seqoff) % PLAYOSIZE)         

        seqin = ntohs(((rtp_hdr_t*)buff)->seq);
        if( seqin < seqoff )
        {
            cout << "e";
            continue;
        }
        
        posin = SEQ2POS(seqin);

        if( seqin > seqinmax )
        {
            seqinmax = seqin;
            posinmax = posin;
        }
        
        bcopy(buff,caroussel[posin].buff,16);
        caroussel[posin].size=n;

        //Markovnak
        AVG_psize+=n;
        AVG_psize_num++;
    }

    //Non blocking socket, check error
    if( errno != EAGAIN )
        cerr << "error retrievind datagram: " << strerror(errno) << endl;

    
    
    
    
    
    
    
    
    //2. Line formating
    if( tsout % 90000 == 0 )
    {

        cout.setf(ios::fixed,ios::floatfield);

        cerr.flush();
        cout << endl;
        
        //Packet statistics
        cout <<GREY<<     "   time=   "<<RESET<<setw(7)<<stat_time;stat_time++;
        cout <<GREY<<      "  pps=    "<<RESET<<setw(7)<<stat_pps;
        cout <<GREY<<      "  lpps=   "<<RESET<<setw(7)<<stat_lpps;
        cout <<GREY<<      "  Bps=    "<<RESET<<setw(7)<<stat_Bps*8/1000;


        cout.precision(5);
        cout <<GREY<<      "  P_Lst=  "<<RESET<<m_out.P_Lsteady;
        cout <<GREY<<      " rP_Lst=  "<<RESET<<m_out.rP_Lsteady;
        cout <<GREY<<      "  P_Rst=  "<<RESET<<m_out.P_Rsteady;   
        cout <<GREY<<      " rP_Rst=  "<<RESET<<m_out.rP_Rsteady;
        
        cout.precision(9);
        //Steady state burst lengths
        for(int i=1;i<5;i++)
            cout <<GREY<<  "  P_Lbu"<<i<<"= "<<RESET<<m_out.P_Lburst[i]; //Maxburst
        cout << "                                                      ";

        for(int i=1;i<5;i++)
            cout <<GREY<<  " rP_Lbu"<<i<<"= "<<RESET<<m_out.rP_Lburst[i]; //Maxburst
        cout << "                                                      ";
        for(int i=1;i<5;i++)
            cout <<GREY<<  "  P_Rbu"<<i<<"= "<<RESET<<m_out.P_Rburst[i]; //Maxburst
        cout << "                                                      ";
        for(int i=1;i<5;i++)
            cout <<GREY<<  " rP_Rbu"<<i<<"= "<<RESET<<m_out.rP_Rburst[i]; //Maxburst

        cout << "                                                      ";

        cout.precision(2);
        cout <<GREY<<      "  reqB_R= "<<RESET<<setw(7)<<m_out.req_B_ret / 1000;
        cout <<GREY<<      "   B_av=  "<<RESET<<setw(7)<<m_in.B_av / 1000;
        static int lastrN_L=0;
        cout <<GREY<<      " delrN_L= "<<RESET<<setw(7)<<m_out.rN_L-lastrN_L;
        lastrN_L=m_out.rN_L;
        
        //Limit req
        cout <<GREY<<      " k_Rgema= "<<RESET<<setw(7)<<m_in.k_Rgeneralmax;
        cout <<GREY<<      " k_Rbuma= "<<RESET<<setw(7)<<m_in.k_Rburstmax;
        cout <<GREY<<      " r_Ribmi= "<<RESET<<setw(7)<<m_out.n_Rinterburstmin;
        
        
        //Rateadapt
        cout <<endl<<"Rateadapt: "<<setprecision(5);
        cout <<GREY<<      " r_20=    "<<RESET<<setw(7)<<m_out.p_20;
        cout <<GREY<<      " r_21=    "<<RESET<<setw(7)<<m_out.p_21;
        
        
        cout <<DOWNUP;
        stat_pps=0;
        stat_Bps=0;
        stat_lpps=0;
        
        //WLAn statistics
        static FILE* fp;
        static char result [1000];
        static int agrCtlRSSI;
        static int agrCtlNoise;
        
        fp = popen("airport -I |grep agrCtlRSSI |sed 's/^.\\{17\\}//g'","r");
        n=fread(result,1,sizeof(result),fp);
        pclose (fp);
        result[n]=0;
        
        sscanf(result,"%d",&agrCtlRSSI); 
        
        fp = popen("airport -I |grep agrCtlNoise |sed 's/^.\\{17\\}//g'","r");
        n=fread(result,1,sizeof(result),fp);
        pclose (fp);
        result[n]=0;
        
        sscanf(result,"%d",&agrCtlNoise); 

        //cout <<GREY<<      " SNR=     "<<RESET<<setw(7)<<agrCtlRSSI  - agrCtlNoise;
        cout <<"    ";
        
        
        //P_L,steady, P_R,steady
        
        
        
        //        cout<<endl;
        cout<<" ";
        
        
        
        //Cerr-nek
        cout.flush();
        cerr << endl;
        cerr << "time= "<<setw(7)<<stat_time-1<<" ";

        
        
        AVG_psize=0;
        AVG_psize_num=0;

    }


#define RATEADAPTx
    
#ifdef RATEADAPT
    //3. Rateadaptation
    static int hit=1;
    if( tsout % 90000 == 0 )
    {
        if( hit % 3 == 0 )
        { 
            //                             300 --> 400
        if( (m_in.B_av>900*1000)&&(m_in.B_ret - m_out.req_B_ret < -100000*1.1) )
        {
            cout << "DOWN"<<endl<<endl;
            n=sendto(sock_cnt,"down\n",5,0,(const struct sockaddr *)&cnt_sockaddr,sizeof(struct sockaddr_in));
            m_in.B_av-=100000;
            m_in.B_ret+=100000;
        }
            //                                  400 --> 300
        if( (m_in.B_av<1300*1000)&&(m_in.B_ret - m_out.req_B_ret >100000*1.1) )
        {
            cout << "UP"<<endl<<endl;
            n=sendto(sock_cnt,"up\n",3,0,(const struct sockaddr *)&cnt_sockaddr,sizeof(struct sockaddr_in));
            m_in.B_av+=100000;
            m_in.B_ret-=100000;
        }
        }
        hit++;
    }
#endif
    
    
    
    

    
    //4. Play out packet(s) (based on timestamp no)
    if( countdown_out==0 )
    {
        if( seqout == -1 )
            seqout=seqoff;
       
        bool hit=false;
        for(int ii=seqout+1,pp=SEQ2POS(ii);ii<seqin;ii++,pp=SEQ2POS(ii))
        {
            //Search packets with tsout 
            if( caroussel[pp].size != 0)
            {
                //Adjust timestamp offset
                if( tsout == -1 )
                    tsout = ntohl(caroussel[pp].prtp->ts);

                if( ntohl(caroussel[pp].prtp->ts) == tsout) 
                {
                    hit=true;
                    //Check for lost packets
                    for(int jj=ii-1;jj>seqout;jj--)
                    {
                        //Packet loss
                        
                        stat_lpps++;
                        
                        markov(LOST,m_in,&m_out, caroussel[SEQ2POS(jj)].retskip);
                        caroussel[SEQ2POS(jj)].size=0;
                        caroussel[SEQ2POS(jj)].retreq=false;
                        caroussel[SEQ2POS(jj)].retskip=false;
                        //cout << "("<<jj<<")";
                    }
                    
                    //Packet match, write out
                    if( caroussel[pp].prtp->ssrc != htonl(0x11111111) )
                    {
                        stat_pps++;
                        stat_Bps+=caroussel[pp].size;
                        markov(GOOD,m_in,&m_out,false);
                    }
                    else
                    {
                        stat_pps++;
                        stat_Bps+=caroussel[pp].size;
                        markov(RET,m_in,&m_out,false);
                    }
                    
                    caroussel[pp].size=0;
                    caroussel[pp].retreq=false;
                    caroussel[pp].retskip=false;
                    
                    seqout=ii;
                }
                
            }

        }
        
        //Update TS
        tsout+=(90000/fps);
    }

    
    
    
#define RET_WINDOW (30*1) //10sec
    
    
    //5. Search missing packets (search missing seq numbers), and request retransmission
    if( countdown_out==0 )
    {
        
#ifdef RET_WINDOW
        
        // General
        static int frame=0;
        static int numrequested[RET_WINDOW];if(frame == 0) for(int i=0;i<RET_WINDOW;i++) numrequested[i]=0;
        static int sumrequested=0;

        numrequested[(frame+1)%RET_WINDOW]=0;
        
        sumrequested=0;
        for(int i=0;i<RET_WINDOW;i++)
            sumrequested+=numrequested[i];
#endif

        
        int end=-1;
        for(int i=seqin-3;i>seqout+3;i--) //reverse direction
        {
            //Search RTP gap end
            if( (caroussel[SEQ2POS(i+1)].size!=0) && (caroussel[SEQ2POS(i)].size==0) )
                end=i;
            
            //Retransmission already requested?
            if( caroussel[SEQ2POS(i)].retreq )
            {
                end=-1;
                continue;
            }
            
            //Search RTP gap start
            if( (caroussel[SEQ2POS(i)].size==0) && (caroussel[SEQ2POS(i-1)].size!=0) && (end != -1))
            {
                
                
                //Lost burst found --> request repair!!!
                for(int j=end,k=1;j+1>i;j--,k++) //reverse direction
                {
                    //Packet loss!!!
                    
                    //Retransmission skip?
                    if( caroussel[SEQ2POS(j)].retskip )
                        continue;
                    
                    
#ifdef RET_WINDOW

                    //1, general rule 
#define k_Rgeneralmax(n)  ((n) * m_in.B_ret / m_in.B_av)

                    if( sumrequested +1 > (frame<RET_WINDOW? k_Rgeneralmax(frame*4):k_Rgeneralmax(RET_WINDOW*4)) )
                    {
                        //Mark for retransmission skip
                        caroussel[SEQ2POS(j)].retskip=true;
                        continue;
                    }
                    
                    //2. burst max
#define k_Rburstmax  ( (m_in.D_playout-2*m_in.D_pkg-m_in.RTT)/(m_in.D_pkg*((m_in.B_av/m_in.B_ret)-1)) );

                    //if( (double)sumrequested*(double)m_in.AVG_psize/(LIMIT_RET/90.0) > m_in.B_ret)
                    //{
                    //    caroussel[SEQ2POS(j)].retskip=true;
                    //    cout << "!";
                    //    continue;
                    //}
                    //1b. In burst limit
                    //if( k>m_in.k_Rburstmax   ) //Reverse direction
                    //{
                    //    //Mark for retransmission skip
                    //    caroussel[SEQ2POS(j)].retskip=true;
                    //    continue;
                    //}
                    numrequested[frame%RET_WINDOW]++;
                    sumrequested++;
#endif
                        
                    char buff[256];
                    sprintf(buff,"get %d",j);
                    //cerr<<j<<buff;
#ifdef TCP
                    if ((ret_sock= socket(AF_INET, SOCK_STREAM, 0)) == -1)
                        cerr<<"error sending command: " << strerror(errno) << endl;   

                    if( connect(ret_sock,(const struct sockaddr *)&ret_sockaddr,sizeof(struct sockaddr_in)) == 0)
                    {
                        if( write(ret_sock,buff,strlen(buff)+1)!=strlen(buff)+1 )
                        {
                            cerr<<"error sending command: " << strerror(errno) << endl;   
                        }
                    }
                    else
                    {
//                       cerr<<"error connecting ret: " << strerror(errno) << endl;   
                    }
                    close(ret_sock);
#else
                    int n=0;
                    n=sendto(ret_sock,buff,strlen(buff),0,(const struct sockaddr *)&ret_sockaddr,sizeof(struct sockaddr_in));
                    if (n != strlen(buff))
                        cerr<<"error sending command: " << strerror(errno) << endl;   
#endif
//                        cerr<<"!";
                    //cerr << "("<<j<<")";
                    caroussel[SEQ2POS(j)].retreq=true;
                }
                
                
            }
            

            //            if( caroussel[i%PLAYOSIZE].size!=0 ) 
            //                seqret = ntohs(caroussel[i%PLAYOSIZE].prtp->seq);
            
            
        }
#ifdef RET_WINDOW
        frame++;
#endif
    }
    else
    {
        countdown_out--;
    }

    
    
    
    
    
    
    return;
}


void markov(unsigned state, markovi_t m_in, markovo_t* m_out, bool retskip)
{
    //0==good
    //1==loss
    //2==ret
    

    static int sw[WINDOW][3][3];
    static unsigned n=0;
    static double s[3][3]={{0,0,0},{0,0,0},{0,0,0}};
//    static unsigned mn=0;
    
    static double rs[3][3]={{1e-13,1e-13,1e-13},{1e-13,1e-13,1e-13},{1e-13,1e-13,1e-13}};
    static unsigned rn=0;
            
    if(state>2)
        return;

    //Display
    switch(state)
    {
        case 0:cerr << ".";break;
        case 1:cerr << (retskip?"s":"x");break;
        case 2:cerr << "r";break;
    };
  
    
    

    
    static unsigned prevstate=0;
    
    //1a. MODEL Number of state transitions

    for(int i=0;i<3;i++)
        for(int j=0;j<3;j++)
            if((i==prevstate)&&(j==state))
                sw[n%WINDOW][i][j]=1;
            else
                sw[n%WINDOW][i][j]=0;
    n++;
//    if( n<WINDOW) mn++;
    
    for(int i=0;i<3;i++)
    {
        for(int j=0;j<3;j++)
        {
            s[i][j]=0;
            for(int k=n-1;(k>0)&&(k>n-WINDOW);k--)
                s[i][j]+=sw[k%WINDOW][i][j];
            s[i][j]+=1e-13;
        }
    }               
    
    //1b. REAL Number of state transitions
    
    rs[prevstate][state]++;
    rn++;


    
    //2. MODEL Probability of state transitions
    static double p[3][3];
    
    p[0][0]=s[0][0]/(s[0][0]+s[0][1]+s[0][2]);
    p[0][1]=s[0][1]/(s[0][0]+s[0][1]+s[0][2]);
    p[0][2]=s[0][2]/(s[0][0]+s[0][1]+s[0][2]);
    p[1][0]=s[1][0]/(s[1][0]+s[1][1]+s[1][2]);
    p[1][1]=s[1][1]/(s[1][0]+s[1][1]+s[1][2]);
    p[1][2]=s[1][2]/(s[1][0]+s[1][1]+s[1][2]);
    p[2][0]=s[2][0]/(s[2][0]+s[2][1]+s[2][2]);
    p[2][1]=s[2][1]/(s[2][0]+s[2][1]+s[2][2]);
    p[2][2]=s[2][2]/(s[2][0]+s[2][1]+s[2][2]);
                
    
    //3a. MODEL Steady states probabilities
    m_out->P_Lsteady=( \
        p[0][2]*p[2][1] + \
        p[0][1]*p[2][0] + \
        p[0][1]*p[2][1] \
        ) / ( \
        p[0][1]*p[2][0] + \
        p[0][1]*p[2][1] + \
        p[0][2]*p[2][1] + \
        p[1][0]*p[0][2] + \
        p[1][2]*p[0][1] + \
        p[1][2]*p[0][2] + \
        p[2][1]*p[1][0] + \
        p[2][0]*p[1][0] + \
        p[2][0]*p[1][2] \
        );

    m_out->P_Rsteady=( \
        p[1][0]*p[0][2] + \
        p[1][2]*p[0][1] + \
        p[1][2]*p[0][2] \
        ) / ( \
        p[0][1]*p[2][0] + \
        p[0][1]*p[2][1] + \
        p[0][2]*p[2][1] + \
        p[1][0]*p[0][2] + \
        p[1][2]*p[0][1] + \
        p[1][2]*p[0][2] + \
        p[2][0]*p[1][0] + \
        p[2][0]*p[1][2] + \
        p[2][1]*p[1][0] \
        );
    
    
    //3b. REAL Steady states probabilities
    m_out->rP_Lsteady = (rs[0][1] + rs[2][1] + rs[1][1])/rn;
    m_out->rP_Rsteady = (rs[0][2] + rs[1][2] + rs[2][2])/rn;
    m_out->rN_L=rs[0][1] + rs[2][1] + rs[1][1];

    
    //4a. MODEL Burst length probabilities
    for(int l=1;l<5;l++)
    {
        m_out->P_Lburst[l]=(p[1][0]+p[1][2]) * pow(1-p[1][0]-p[1][2],l-1);
        m_out->P_Rburst[l]=(p[2][0]+p[2][1]) * pow(1-p[2][0]-p[2][1],l-1);
    }
    
    
    //4b. REAL Burst length probabilities
    static int start=-1;
    static int Lburst[10]={0,0,0,0,0,0,0,0,0,0};
    static int Rburst[10]={0,0,0,0,0,0,0,0,0,0};
    static int Lburstnum=0;
    static int Rburstnum=0;
    
    
    //Start burst
    if( ((prevstate==0)&&(state==1)) || ((prevstate==2)&&(state==1)) )
        start=n;
    if( ((prevstate==0)&&(state==2)) || ((prevstate==1)&&(state==2)) )
        start=n;
       
    //end burst
    if( ((prevstate==1)&&(state==0)) || ((prevstate==1)&&(state==2)) )
    {
        Lburst[n-start]++;
        Lburstnum++;
    }
    if( ((prevstate==2)&&(state==0)) || ((prevstate==2)&&(state==1)) )
    {
        Rburst[n-start]++;
        Rburstnum++;
    }

    m_out->rP_Lburst[1]=Lburst[1]/(Lburstnum+1e-13);
    m_out->rP_Lburst[2]=Lburst[2]/(Lburstnum+1e-13);
    m_out->rP_Lburst[3]=Lburst[3]/(Lburstnum+1e-13);
    m_out->rP_Lburst[4]=Lburst[4]/(Lburstnum+1e-13);

    m_out->rP_Rburst[1]=Rburst[1]/(Rburstnum+1e-13);
    m_out->rP_Rburst[2]=Rburst[2]/(Rburstnum+1e-13);
    m_out->rP_Rburst[3]=Rburst[3]/(Rburstnum+1e-13);
    m_out->rP_Rburst[4]=Rburst[4]/(Rburstnum+1e-13);


    
    
    //3.3.2 retransmission limitation
    

    
    
        
    
    //3.3.3 Rateadapt
    //p[i][j] --> P_Rsteady --> req_B_ret
    //m_out->req_B_ret = m_in.B_av * m_out->P_Rsteady;
    
    //m_out->p_20=p[2][0];
    //m_out->p_21=p[2][1];
    
    
    
    
    
    
    //Valami
//    m_out->S_burst    =( (m_in.D_playout-2*m_in.D_pkg-m_in.RTT)/(1/m_in.B_ret-1/m_in.B_av) ); //byte
//    m_out->k          =( (m_in.D_playout-2*m_in.D_pkg-m_in.RTT)/m_in.D_ret );                 //piece
    
    
    
    
    prevstate=state;

    return;


    
}