//
//  ret.cpp
//  iptv
//
//  Created by Tamás Jursonovics on 2012.05.13..
//  Copyright (c) 2012 Tamás Jursonovics. All rights reserved.
//
//  Captures the multicast traffic in a circular buffer and implements a simple retransmission request protocol.
//

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>
#include "rtp.h"
#include "nal.h"
#include <errno.h>
#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

using namespace std;


//Receiver thread
void *th_recv(void *arg);
struct recv_arg_t{
    char src_addr[256];
    int src_port;
    char src_intf[256];
};
recv_arg_t recv_arg;


//Responder thread
void *th_resp(void *arg);
struct resp_arg_t{
    char src_addr[256];
    int src_port;
    int cnt_port;
};
resp_arg_t resp_arg;

#define FRAMEBUFFER (30*10) //10sec by 30fps

//Caroussel for storing incomming packets
#define MAXPACKET 200000    //Max size of a packet.
struct playout_t{
    char buff[MAXPACKET];
    int size;
    rtp_hdr_t *prtp;
};

playout_t caroussel[FRAMEBUFFER];
pthread_mutex_t m_playo;


bool stop=false;




int main (int argc, const char * argv[])
{
    cout << "#Simple UDP/RTP retransmitter" << endl<<"#"<<endl;
    cerr.flush();
    if (argc!=5)
    {
        cout << "Usage: "<<argv[0]<<" <mcs_addr> <mcs_port> <mcs_intf> <cnt_port>"<<endl<<endl;
        cout << "Listens for multicast from <mcs_addr>:<mcs_port> on <mcs_intf> interface. Control port on <cnt_port>."<<endl;
        cout << "Commands: \"get <xx>\": resends the <xx> RTP packet to <mcs_addr>:<mcs_port>"<<endl;
        cout << "          \"bye\"     : exit"<<endl<<endl;
        cout << "(you may use nc utility for a human intercafe)"<<endl<<endl;
        cout << "Retransmitted packets are marked with SSRC=0x11111111."<<endl<<endl;
        
        return 0;
    }

    strcpy(recv_arg.src_addr,argv[1]);
    recv_arg.src_port=atoi(argv[2]);
    strcpy(recv_arg.src_intf,argv[3]);
    resp_arg.cnt_port=atoi(argv[4]);
    cout << "#mcs_addr = "<<recv_arg.src_addr<<endl;
    cout << "#mcs_port = "<<recv_arg.src_port<<endl;
    cout << "#mcs_intf = "<<recv_arg.src_intf<<endl;
    cout << "#cnt_port = "<<resp_arg.cnt_port<<endl;
    cout << "#framebuf = "<<FRAMEBUFFER<<" frame"<<endl;
    cout << "SSRC_mark = 0x11111111"<<endl<<"#"<<endl;
    
    
    //Initiallize playout buffer
    for(int i=0;i<FRAMEBUFFER;i++)
    {
        caroussel[i].size=0;
        caroussel[i].prtp=(rtp_hdr_t*)caroussel[i].buff;
        //Fu header
    }
    if( pthread_mutex_init(&m_playo,NULL) != 0)
    {
        cout<<"Error creatig mutex: " << strerror(errno)<<endl;
        return -1;
    }
    
    
    //Start receiver thread
    pthread_t threcv;
    if( pthread_create(&threcv, NULL, th_recv, &recv_arg) )
    {
        cout<<"Error creatig thread: " << strerror(errno)<<endl;
        pthread_mutex_destroy(&m_playo);
        return -1;
    }

    
    //Start sender responder thread
    strcpy(resp_arg.src_addr,argv[1]);
    resp_arg.src_port=atoi(argv[2]);
    pthread_t thresp;
    if( pthread_create(&thresp, NULL, th_resp, &resp_arg) )
    {
        cout<<"Error creatig thread: " << strerror(errno)<<endl;
        pthread_mutex_destroy(&m_playo);
        pthread_cancel(threcv);
        return -1;
    }
    
    
    //threads are working...
    
    //wait for responder thread end
    if( (pthread_join(thresp,NULL)) != 0 )
    {
        cout<<"Error joining thread: " << strerror(errno)<<endl;
        pthread_mutex_destroy(&m_playo);
        pthread_cancel(threcv);
        pthread_cancel(thresp);
        return -1;
    }
    
    //wait for receiver thread end
    if( (pthread_join(threcv,NULL)) != 0 )
    {
        cout<<"Error joining thread: " << strerror(errno)<<endl;
        pthread_mutex_destroy(&m_playo);
        pthread_cancel(threcv);
        return -1;
    }

    pthread_mutex_destroy(&m_playo);
    
    cout << "#byebye"<<endl;
    
    return 0;
}




void *th_recv(void *arg)
{
    recv_arg_t *precv_arg=(recv_arg_t*)arg;
    
    
    //initialize receive socket
    struct sockaddr_in mcast;
    int mlen=sizeof(mcast);
    struct ip_mreq group;
    int rsock, n;
    
    //Create a datagram socket on which to receive.
    if( (rsock=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"Error creating socket: " << strerror(errno)<<endl;
        stop=true;
        return NULL;
    }
    
    // Enable SO_REUSEADDR to allow multiple instances of this application to receive copies of the multicast datagrams.
    int reuse = 1;
    if(setsockopt(rsock, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse, sizeof(reuse)) < 0)
    {
        cout<<"Error reuse socket: " << strerror(errno)<<endl;
        close(rsock);
        stop=true;
        return NULL;
    }
    
    // Bind to the proper port number with the IP address
    memset((char *) &mcast, 0, sizeof(mcast));
    mcast.sin_family = AF_INET;
    mcast.sin_addr.s_addr = inet_addr(precv_arg->src_addr);
    mcast.sin_port = htons(precv_arg->src_port);
    if(bind(rsock, (struct sockaddr*)&mcast, sizeof(mcast)))
    {
        cout<<"Error bind socket: " << strerror(errno)<<endl;
        close(rsock);
        stop=true;
        return NULL;
    } 
    
    // Join the multicast group on the local
    group.imr_multiaddr.s_addr = inet_addr(precv_arg->src_addr);
    group.imr_interface.s_addr = inet_addr(precv_arg->src_intf);
    if(setsockopt(rsock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&group, sizeof(group)) < 0)
    {
        cout<<"Error mcast join socket: " << strerror(errno)<<endl;
        close(rsock);
        stop=true;
        return NULL;
    } 
    
    
    //Initialize buffer
    char buff[MAXPACKET];
    static unsigned cwrite=0;
    while(!stop)
    {
        /* Read from the socket. */
        if( (n=recvfrom(rsock, buff, sizeof(buff),0,(struct sockaddr*)&mcast,(socklen_t *)&mlen)) < 0 )
        {
            cout << "Error retrievind datagram: " << strerror(errno) << endl;
        }
        else 
        {
            //Size?
            if( n<sizeof(rtp_hdr_t) )
            {
                cout <<"#Malformatted RTP packet: too small packet\n"<<endl;
                continue;
            }
            
            if( n>MAXPACKET )
            {
                cout << "#Malformatted RTP packet: oversized packet\n"<<endl;
                continue;
            }
            
            //Version?
            if( ((rtp_hdr_t*)buff)->version != 2 )
            {
                cout << "#Malformatted RTP packet: unknown RTP version\n"<<endl;
                continue;
            }
            
            //Put into the caroussel in arriving order
            pthread_mutex_lock(&m_playo);
            
            bcopy(buff,caroussel[cwrite].buff,n);
            caroussel[cwrite].size=n;
            
            pthread_mutex_unlock(&m_playo);

            cwrite++;
            if( cwrite == FRAMEBUFFER )
                cwrite=0;


            cerr << ".";
        }
        
    }
    
    close(rsock);
    
    return NULL;
}

void *th_resp(void *arg)
{
    resp_arg_t *presp_arg=(resp_arg_t*)arg;


//Comment out to use UDP for control socket
#define TCP

    
    //initialize control socket
    struct sockaddr_in addr;
    socklen_t addr_len;
    int csock, addrlen=sizeof(addr);

#ifdef TCP
    if( (csock=socket(AF_INET, SOCK_STREAM, 0)) == -1)
#else
    if( (csock=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
#endif
    {
        cout<<"Error creating socket: " << strerror(errno)<<endl;
        stop=true;
        return NULL;
    }
    
    memset(&addr,0,sizeof(addr));
    addr.sin_family=AF_INET;
    addr.sin_addr.s_addr=htonl(INADDR_ANY); /* N.B.: differs from sender */
    addr.sin_port=htons(presp_arg->cnt_port);
    
    if( bind(csock,(struct sockaddr *) &addr,sizeof(addr)) != 0)
    {
        cout << "Error binding control socket: " << strerror(errno)<<endl;
        close(csock);
        stop=true;
        return NULL;
    }
    
    
#ifdef TCP
    //SO_KEEPALIVE
    int optval = 1;
    socklen_t optlen = sizeof(optval);

    optlen = sizeof(optval);
    if(setsockopt(csock, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen) < 0) {
        cout << "Error setsockopt socket: " << strerror(errno)<<endl;
        close(csock);
        stop=true;
        return NULL;
    }
    if( listen(csock,1) == -1 )
    {
        cout << "Error listening control socket: " << strerror(errno)<<endl;
        close(csock);
        stop=true;
        return NULL;
    }
#endif
        
    //Initialize sender socket (socket for RTP packet retransmission)
    struct sockaddr_in dst;
    int ssock;
    struct hostent *hp;

    if( (ssock=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cout<<"Error creating server socket: " << strerror(errno)<<endl;
        close(csock);
        stop=true;
        return NULL;
    }
    
    dst.sin_family = AF_INET;
    if ((hp = gethostbyname(presp_arg->src_addr)) == 0)
    {
        cout<<"Error: unknown host: " << presp_arg->src_addr << strerror(errno)<<endl;
        close(ssock);
        close(csock);
        stop=true;
        return NULL;
    }
    
    bcopy((char *)hp->h_addr, 
          (char *)&dst.sin_addr,
          hp->h_length);
    dst.sin_port = htons(presp_arg->src_port);

    
    char buff[MAXPACKET];
    int n;
#ifdef TCP
    int conn;
    cout << "#waiting for conn"<<endl;
    
    int stat_p=0; //packet
    int stat_B=0; //byte
    double stat_now=0, stat_last=0;
    struct timeval stat_time;

    
    while( (!stop) && ((conn=accept(csock,(struct sockaddr *)&addr,&addr_len)) >= 0) )
#else
    while( !stop )
#endif
    {
#ifdef TCP
        while( (!stop) && ((n=read(conn,buff,255)) >= 0) )
#else
        while( (n=recvfrom(csock,buff,255,0,(struct sockaddr *)&addr,&addr_len)) >= 0 )
#endif
        {

            //Parse packet
            if( strncmp("get ",buff,4) == 0 )
            {
                //Parse RTP seq no
                u_int16 seq=-1;
                if( sscanf(buff,"get %u",&seq) != 1)
                {
                    cout << "#Invalid get command: "<<buff<<endl;
                    close(conn);
                    continue;
                }
                    
                //Search in caroussel
                int i=0;
                pthread_mutex_lock(&m_playo);
                
                while( i<FRAMEBUFFER )
                {
                    if( ntohs(caroussel[i].prtp->seq) == seq )
                        break;
                    
                    i++;
                }
                bcopy(caroussel[i%FRAMEBUFFER].buff,buff,caroussel[i%FRAMEBUFFER].size);
                n=caroussel[i%FRAMEBUFFER].size;

                pthread_mutex_unlock(&m_playo);
                
                //hit?
                if( i != FRAMEBUFFER)
                {
                    //Mark packet for markov analysis --> It is easier for the client to differenciate packets originated from the server and retransmitter. The client does not need to parse L3 headers.
                    ((rtp_hdr_t*)buff)->ssrc=htonl(0x11111111);
                    
                    //Send resp
                    if( sendto(ssock,buff,n,0,(const struct sockaddr *)&dst,sizeof(struct sockaddr_in)) != n )
                        cout<<"Error sending datagram: " << strerror(errno) << endl;
                    
                    stat_p++;
                    stat_B+=n;
                    cerr << "("<<ntohs(((rtp_hdr_t*)buff)->seq)<<")";
                }
                else
                {
                    cerr<<"(?)";
                }
            }
            else if( strncmp("bye",buff,3) == 0 )
            {
                stop=true;
            }
            else
            {
                cout <<"#Unknown command: "<<buff<<endl;
            }
            close(conn);
            
            //Stats
            gettimeofday(&stat_time, NULL);
            stat_now = (double)stat_time.tv_sec+(double)stat_time.tv_usec/1000000;
            if(  stat_now > stat_last + 1 )
            {
                cout.precision(2);
                cout.setf(ios::fixed,ios::floatfield);
                static double offset = stat_now;
                cout << endl <<setw(6)<<stat_now-offset<<" s, " <<setw(3)<< stat_p <<" packet, ";
                cout.precision(0);
                cout.setf(ios::fixed,ios::floatfield);
                cout << setw(4)<<(double)stat_B*8/(stat_now-stat_last)/1000 <<" kbps";

                stat_B=0;
                stat_p=0;

                stat_last=stat_now;
            }
                
            
        }
              
    }
    
    //Destroy socket
    cout <<"#Connection closed"<<endl;
    close(csock);
    close(ssock);
    
    return NULL;
}

