//  Szerintem ez rossz
//
//  ser.cpp
//  iptv
//
//  Created by Tomi on 2012.05.13..
//  Copyright (c) 2013 Tamás Jursonovics. All rights reserved.
//
//  A simple H.264 packetizer, generates UDP/RTP multicast traffic based on the RFC 6184 and implements a simple control protocol.
//


#include <iostream>
#include <fstream>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include "rtp.h"
#include "nal.h"
#include <semaphore.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <pthread.h>


#ifdef _WIN32 // note the underscore: without it, it's not msdn official!
// Windows (x64 and x86)
#elif __unix__ // all unices
// Unix
#include <cstdlib>
#include <string.h>
#include <fcntl.h>
#elif __posix__
// POSIX
#elif __linux__
// linux
#elif __APPLE__
// Mac OS, not sure if this is covered by __posix__ and/or __unix__ though...
#endif


#define HEAD_START_CODE    0x010000  //little endian (reverse order for Intel)
#define PICTURE_START_CODE 0x00
#define SLICE_MIN_START    0x01
#define SLICE_MAX_START    0xAF
#define USER_START_CODE    0xB2
#define SEQ_START_CODE     0xB3
#define EXT_START_CODE     0xB5
#define SEQ_END_CODE       0xB7
#define GOP_START_CODE     0xB8


using namespace std;


sem_t *sem;
static int fps;
void *thread(void *arg);
void timer_handler (int signum);


unsigned long getnal(ifstream &mpeg_input, unsigned char* RBSP_data);
unsigned int expGolomb_UD(unsigned int *bitcounter, unsigned char* RBSP_data);
bool getRawBit(unsigned int *bitcounter, unsigned char* nal);


int main (int argc, const char * argv[])
{
    cerr.flush();
    if (argc!=5)
    {
        cout << "Simple UDP/RTP h264 packetizer" << endl<<endl;
        //                              1         2          3         4
        cout<<"Usage: "<<argv[0]<<" <h264_file> <dst_addr> <dst_port> <fps>"<<endl<<endl;
        cout<<"Dummy UDP/RTP packetizes. Listen for commands on <lst_port> and streams back to the port from which the hello command arrived with <fps> rate."<<endl;
        cout << "Commands: \"hello\" : start"<<endl;
        cout << "          \"bye\"   : exit"<<endl;
        //        cout << "          \"up\"    : switch to next file"<<endl;
        //        cout << "          \"down\"  : switch to previous file"<<endl;
        cout << "(you may use nc utility for a human intercafe)"<<endl;

        return 0;
    }
    cerr << "Running program ="<<argv[0]<<"\n";
    cerr << "h264_file = "<<argv[1]<<endl;
    cerr << "dst_addr  = "<<argv[2]<<endl;
    cerr << "dst_port = "<<argv[3]<<endl;
    cerr << "fps  = "<<argv[4]<<endl<<endl;
    
    //Create semaphore
    char name[256];
    srand( time(NULL) );
    sprintf(name,"ser_sem_%d",rand());
    if( (sem=sem_open(name, O_CREAT, 0,0)) == SEM_FAILED )
    {
        cout << "Error creating semaphore: " << strerror(errno) <<endl;
        return -1;
    }
        
    //Initialie file
    ifstream in_stream(argv[1], ifstream::in);
    if (in_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[1] << ":" << strerror(errno);
        sem_close(sem);
        return -1;
    }
    
    unsigned char* rtp = (unsigned char*)malloc((size_t)sizeof(char)*100000);
    unsigned char* nalp = &rtp[sizeof(rtp_hdr_t)];
    int nals;
    unsigned int bitcounter=0;
    
    unsigned int first_mb_in_slice=-1;
    unsigned int slice_type=-1;
    
    //initialize socket
    int sock, n;
    struct sockaddr_in dst;
    struct hostent *hp;
    
    if ((sock= socket(AF_INET, SOCK_DGRAM, 0)) == -1)
    {
        cerr<<"error creating socket: " << strerror(errno);
        sem_close(sem);
        in_stream.close();
        free(rtp);
        return -1;
    }
    
    int buffsize=25000;
    int size=sizeof(buffsize);
    if( setsockopt(sock,SOL_SOCKET,SO_SNDBUF,&buffsize,size) == -1 )
    {
        cerr<<"error setsockopt: " << strerror(errno);
        sem_close(sem);
        in_stream.close();
        free(rtp);
        close(sock);
        return -1;
    }    
    
    dst.sin_family = AF_INET;
    if ((hp = gethostbyname(argv[2])) == 0)
    {
        cerr<<"Unknown host: " << argv[2] << strerror(errno);
        sem_close(sem);
        in_stream.close();
        free(rtp);
        close(sock);
        return -1;
    }
    
    bcopy((char *)hp->h_addr, 
          (char *)&dst.sin_addr,
          hp->h_length);
    dst.sin_port = htons(atoi(argv[3]));
    
    
    //Parse fps
    sscanf(argv[4],"%d",&fps);
    
    
    //Init RTP header
    ((rtp_hdr_t*)rtp)->version=2;
    ((rtp_hdr_t*)rtp)->p=0;
    ((rtp_hdr_t*)rtp)->x=0;
    ((rtp_hdr_t*)rtp)->cc=1;
    ((rtp_hdr_t*)rtp)->m=0;
    ((rtp_hdr_t*)rtp)->pt=96;
    ((rtp_hdr_t*)rtp)->seq=htons(0);//(srand(time(NULL)));
    ((rtp_hdr_t*)rtp)->ts=0;
    ((rtp_hdr_t*)rtp)->ssrc=htonl(0x12345678);
    ((rtp_hdr_t*)rtp)->csrc[0]=htonl(0x87654321);
    
    
    //Create thread
    pthread_t mythread;
    if( pthread_create(&mythread, NULL, thread, NULL) )
    {
        cout<<"Error creatig thread: " << strerror(errno);
        sem_close(sem);
        in_stream.close();
        free(rtp);
        close(sock);
        return -1;
    }

    //Start cycle    
    while (in_stream.good())
    {
        /*
         1. Search HEAD_START_CODE, gen NAL
         [00] 00 00 01
         */
        
        nals=getnal(in_stream, nalp);
        
        //Biztos hogy az egész NAL-t be kell tenni az RTP headerbe? Ennek utánanétni az RFC-ben!!!
        
        /*         
         2. parse NAL UNIT HEADER, check type
         +---------------+
         |0|1|2|3|4|5|6|7|
         +-+-+-+-+-+-+-+-+
         |F|NRI|  Type   |
         +---------------+
         */
        
        if( ((nalp[0] & 0x1F) > 0) && ((nalp[0] & 0x1F) < 24))
        {
            switch (nalp[0] & 0x1F)
            {
                case 1: //Coded slice of a non-IDR picture
                case 5: //Coded slice of an IDR picture
                    
                    /*
                     3. parse SLICE HEADER, check slice_type
                     
                     slice_layer_without_partitioning_rbsp( ) {
                     slice_header( )
                     slice_data( ) // all categories of slice_data( ) syntax 
                     rbsp_slice_trailing_bits( )
                     }
                     
                     slice_header( ) {
                     first_mb_in_slice         2 ue(v)
                     slice_type                2 ue(v)
                     pic_parameter_set_id      2 ue(v)
                     frame_num                 2 u(v)
                     if( !frame_mbs_only_flag ) {
                     field_pic_flag        2 u(1)
                     if( field_pic_flag )
                     bottom_field_flag 2 u(1)
                     }
                     if( nal_unit_type = = 5 )
                     idr_pic_id            2 ue(v)
                     if( pic_order_cnt_type = = 0 ) {
                     pic_order_cnt_lsb 2 u(v)
                     if( pic_order_present_flag && !field_pic_flag )
                     delta_pic_order_cnt_bottom 2 se(v)
                     }
                     }
                     */
                    
                    bitcounter=0;
                    
                    //skip first parameter coded by expGolomb
                    first_mb_in_slice=expGolomb_UD(&bitcounter,&nalp[1]);
                    
                    //decode second parameter (slice_type)
                    slice_type=expGolomb_UD(&bitcounter,&nalp[1]);
                    
                    //IDR / NON-IDR
                    if ((nalp[0] & 0x1F) == 1 )
                    {
                        cout<<" ";
                        ((rtp_hdr_t*)rtp)->m=0;
                    }
                    else
                    {
                        cout<<"i";
                        ((rtp_hdr_t*)rtp)->m=1;
                    }
                    switch (slice_type)
                    {
                        case 0: cout<<"P ";break;
                        case 1: cout<<"B ";break;
                        case 2: cout<<"I ";break;
                        case 3: cout<<"SP";break;
                        case 4: cout<<"SI";break;
                        case 5: cout<<"P ";break;
                        case 6: cout<<"B ";break;
                        case 7: cout<<"I ";break;
                        case 8: cout<<"SP";break;
                        case 9: cout<<"SI";break;
                        default: cout<<"?"<<slice_type<<"?";
                    }
                    
                    break;
                case 2:
                    cout << "Coded slice data partition A slice_data_partition_a_layer_rbsp ";break;
                case 3:
                    cout << "Coded slice data partition B slice_data_partition_b_layer_rbsp ";break;
                case 4:
                    cout << "Coded slice data partition C slice_data_partition_c_layer_rbsp ";break;
                case 6:
                    cout << "SEI ";break;
                case 7:
                    cout << "Sequence parameter set ";break;
                case 8:
                    cout << "Picture parameter set ";break;
                case 9:
                    cout << "Access unit delimiter access_unit_delimiter_rbsp ";break;
                case 10:
                    cout << "End of sequence end_of_seq_rbsp ";break;
                case 11:
                    cout << "End of stream end_of_stream_rbsp ";break;
                case 12:
                    cout << "Filler data filler_data_rbsp ";break;
                default:
                    cout << "Unknown(";
            }
            cout << " " << nals << "B";
            
        }

        
        //4. Wait for semaphore
        sem_wait(sem);
   
        
        /*
         5. Send NAL
         
         Single NAL Unit Packet: Contains only a single NAL unit in the
         payload.  The NAL header type field will be equal to the original NAL
         unit type; i.e., in the range of 1 to 23, inclusive.  Specified in
         section 5.6.
         
         Table 1.  Summary of NAL unit types and their payload structures
         
         Type   Packet    Type name                        Section
         ---------------------------------------------------------
         0      undefined                                    -
         1-23   NAL unit  Single NAL unit packet per H.264   5.6
         24     STAP-A    Single-time aggregation packet     5.7.1
         25     STAP-B    Single-time aggregation packet     5.7.1
         26     MTAP16    Multi-time aggregation packet      5.7.2
         27     MTAP24    Multi-time aggregation packet      5.7.2
         28     FU-A      Fragmentation unit                 5.8
         29     FU-B      Fragmentation unit                 5.8
         30-31  undefined                                    -
         */
        
        ((rtp_hdr_t*)rtp)->ts=htonl(ntohl(((rtp_hdr_t*)rtp)->ts)+90000/fps);
                
        //Kell fragmentáció?
        int MNU = 1400-sizeof(rtp_hdr_t);
        if( nals < MNU)
        {
            n=sendto(sock,rtp,sizeof(rtp_hdr_t)+nals,0,(const struct sockaddr *)&dst,sizeof(struct sockaddr_in));
            if (n != sizeof(rtp_hdr_t)+nals) cerr<<"error sending datagram: " << strerror(errno) << endl;
            
            cout << " " << ntohs(((rtp_hdr_t*)rtp)->seq) << endl;
            
            ((rtp_hdr_t*)rtp)->seq=htons(ntohs(((rtp_hdr_t*)rtp)->seq)+1);
        }
        else
        {
            //0. Create temporary location
            unsigned char buff[1500];
        

            /*1. Create FU indicator and header 
                FU indicator
                
                +---------------+
                |0|1|2|3|4|5|6|7|
                +-+-+-+-+-+-+-+-+
                |F|NRI|  Type   |
                +---------------+
             */
            buff[sizeof(rtp_hdr_t)]=28;
            
            /*3. The FU header has the following format:
                
                +---------------+
                |0|1|2|3|4|5|6|7|
                +-+-+-+-+-+-+-+-+
                |S|E|R|  Type   |
                +---------------+
                */
            
            fu_hdr_t fu_hdr;
            fu_hdr.s=1;
            fu_hdr.e=0;
            fu_hdr.r=0;
            fu_hdr.type=13;
            
            while(nals>0)
            {  
                //Copy RTP header
                bcopy(rtp,buff,sizeof(rtp_hdr_t));

                //Last FU header?
                if( nals<MNU )
                    fu_hdr.e=1;
                
                bcopy(&fu_hdr,&buff[sizeof(rtp_hdr_t)+1],1);

                //First FU header?
                if(fu_hdr.s==1)
                    fu_hdr.s=0;
                
                //Payload
                bcopy((rtp+sizeof(rtp_hdr_t)),&buff[sizeof(rtp_hdr_t)+2],MNU-2);
                
                //3.b Send
                n=sendto(sock,buff,sizeof(rtp_hdr_t)+((nals<MNU-2)?nals:MNU-2),0,(const struct sockaddr *)&dst,sizeof(struct sockaddr_in));
                if (n != sizeof(rtp_hdr_t)+((nals<MNU-2)?nals:MNU-2)) cerr<<"error sending datagram: " << strerror(errno) << endl;

                cout << " " << ntohs(((rtp_hdr_t*)rtp)->seq);
                
                nals-=MNU-2;
                ((rtp_hdr_t*)rtp)->seq=htons(ntohs(((rtp_hdr_t*)rtp)->seq)+1);
                
            }
            cout << endl;
        }
        
        
        //ugy ahogy van, belerakom a socketbe... :)
        //Create RTP packet...
        
                      
        //        n=sendto(sock,"Tomi",strlen("Tomi"),0,(const struct sockaddr *)&dst,sizeof(struct sockaddr_in));
        //        n=sendto(sock,rtp,size+4*4,0,(const struct sockaddr *)&dst,sizeof(struct sockaddr_in));
        //            if (n < 0) cerr<<"error sending datagram: " << strerror(errno) << endl;
        
        

        
    }
    
    
    sem_close(sem);
    in_stream.close();
    free(rtp);
    close(sock);
    
    pthread_cancel(mythread);

    return 0;
}

//http://www.ietf.org/rfc/rfc3984.txt


void timer_handler (int signum)
{
    //Increment semaphore
    sem_post(sem);
}


unsigned long getnal(ifstream & mpeg_input, unsigned char* nal)
{
    static unsigned long head_value=0xffffffff;
    unsigned char * a;
    a=(unsigned char *)&head_value;
    
    
    unsigned int pos=-1;
    
    while (mpeg_input)
    {   
        //NAL start?
        if ((head_value&0x00FFFFFF)==HEAD_START_CODE)
        {
            if( pos == -1 )
            {
                //First nal, save nal byte
                pos=0;
            }
            else
            {
                //Second nal, exit
                return pos-1;
            }
        }
        
        if( pos != -1)
        {
            //record nal
            nal[pos]=a[3];
            pos++;
        }   

        //Read byte
        head_value=head_value>>8;
        a[3]=mpeg_input.get();
        
    }
    
    // End of file reached
    cout<<" End of broked file at "<<pos<<endl;
    return pos;
}

//Decoder functions

unsigned int expGolomb_UD(unsigned int *bitcounter, unsigned char* RBSP_data)
{
    unsigned int M = 0;
    unsigned int info = 0;
    bool bit = 0;
    unsigned int val = 0;
    
    do
    {
        bit = getRawBit(bitcounter,RBSP_data);
        if ( !bit )
            M++;
    }
    while( !bit && M<64 );//terminate if the number is too big!
    
    //now get the M info bits       
    for ( unsigned int i=0 ; i<M ; ++i)
    {
        bit = getRawBit(bitcounter,RBSP_data);
        if ( bit==1 ) 
        {
            //1 jött
            info |= (1<<i);
        }
        else
        {
            //0 jött
            info=info<<1;
        }
    }// i    
    val = (1<<M) + info;
    val -=1;
    
    return val;
}

bool getRawBit(unsigned int *bitcounter, unsigned char* RBSP_data)
{
    unsigned int returnValue=0;
    unsigned int count=0;
    unsigned char RBSP_current_byte=0;
    while(count<=(*bitcounter))
    {
        //8-byte fast forward 
        if (((*bitcounter)-count)>7)
        {
            count+=8;
            RBSP_current_byte++;
            RBSP_data=&RBSP_data[RBSP_current_byte];
            continue;
        }
        
        //Classic bit by bit loading
        returnValue=((RBSP_data[RBSP_current_byte]>>(7-(*bitcounter)))&1);
        count++;
    }
    
    (*bitcounter)++;   
    return returnValue;
}


void *thread(void *arg)
{
   //Init timer
    struct sigaction sa;
    struct itimerval timer;
    
    /* Install timer_handler as the signal handler for SIGVTALRM. */
    memset (&sa, 0, sizeof (sa));
    sa.sa_handler = &timer_handler;
    sigaction (SIGVTALRM, &sa, NULL);
    
    /* Configure the timer to expire*/
    timer.it_value.tv_sec = 0;
    timer.it_value.tv_usec = 1000000/fps;
    /* ... and after that. */
    timer.it_interval.tv_sec = 0;
    timer.it_interval.tv_usec = 1000000/fps;
    /* Start a virtual timer. It counts down whenever this process is
     executing. */
    if (setitimer (ITIMER_VIRTUAL, &timer, NULL) != 0 )
    {
        cerr << "error stting timer: " << strerror(errno);
        
        //pthread_exit(mythread);
        sem_close(sem);
        return NULL;
    }

    //wait
    while(1);
    
    return NULL;
}
