//
//  ana.cpp
//  iptv tools
//
//  Created by Tamas Jursonovics on 2012.06.30..
//  Copyright (c) 2012 Tamás Jursonovics. All rights reserved.
//
//  A simple h264 NALU analyser.
//


#include <iostream>
#include <iomanip>
#include <fstream>
#include <errno.h>
#include <stdio.h>

#include "rtp.h"
#include "nal.h"
#include "iptv.h"


using namespace std;


int main(int argc, const char * argv[])
{
    if (argc!=2)
    {
        std::cout<<"Usage: "<<argv[0]<<" <h264_video_file>"<<endl<<endl;
        return 0;
    }
    
    cout << "#Simple h264 analyser" << endl<< "#"<<endl;
    cout << "#idr             {i,_}" << endl;
    cout << "#size            [byte]" << endl;
    cerr.flush();
    cerr << "#Running program "<<argv[0]<<"\n";
    cerr << "#file            "<<argv[1]<<endl<<"#"<<endl;
    cout << "# No idr type     size" << endl;
    
    //Initialie file
    ifstream in_stream(argv[1], ifstream::in);
    if (in_stream.fail() )
    {
        cerr<<"Error opening the file " << argv[1] << ":" << strerror(errno);
        return -1;
    }
    
    unsigned char* rtp = (unsigned char*)malloc((size_t)sizeof(char)*100000);
    unsigned char* nalp = &rtp[sizeof(rtp_hdr_t)];
    int nals;
    unsigned int bitcounter=0;
    
    unsigned int first_mb_in_slice=-1;
    unsigned int slice_type=-1;
    unsigned int nal=0;
    
    
    //Start cycle    
    while (in_stream.good())
    {
        /*
         1. Search HEAD_START_CODE, gen NAL
         [00] 00 00 01
         */
        
        nals=getnal(in_stream, nalp);
                
        /*         
         2. parse NAL UNIT HEADER, check type
         +---------------+
         |0|1|2|3|4|5|6|7|
         +-+-+-+-+-+-+-+-+
         |F|NRI|  Type   |
         +---------------+
         */
        
        cout << setw(4) << nal << " ";nal++;
        
        if( ((nalp[0] & 0x1F) > 0) && ((nalp[0] & 0x1F) < 24))
        {
            switch (nalp[0] & 0x1F)
            {
                case 1: //Coded slice of a non-IDR picture
                case 5: //Coded slice of an IDR picture
                    
                    /*
                     3. parse SLICE HEADER, check slice_type
                     
                     slice_layer_without_partitioning_rbsp( ) {
                     slice_header( )
                     slice_data( ) // all categories of slice_data( ) syntax 
                     rbsp_slice_trailing_bits( )
                     }
                     
                     slice_header( ) {
                     first_mb_in_slice         2 ue(v)
                     slice_type                2 ue(v)
                     pic_parameter_set_id      2 ue(v)
                     frame_num                 2 u(v)
                     if( !frame_mbs_only_flag ) {
                     field_pic_flag        2 u(1)
                     if( field_pic_flag )
                     bottom_field_flag 2 u(1)
                     }
                     if( nal_unit_type = = 5 )
                     idr_pic_id            2 ue(v)
                     if( pic_order_cnt_type = = 0 ) {
                     pic_order_cnt_lsb 2 u(v)
                     if( pic_order_present_flag && !field_pic_flag )
                     delta_pic_order_cnt_bottom 2 se(v)
                     }
                     }
                     */
                    
                    bitcounter=0;
                    
                    //skip first parameter coded by expGolomb
                    first_mb_in_slice=expGolomb_UD(&bitcounter,&nalp[1]);
                    
                    //decode second parameter (slice_type)
                    slice_type=expGolomb_UD(&bitcounter,&nalp[1]);
                    
                    //IDR / NON-IDR
                    if ((nalp[0] & 0x1F) == 1 )
                    {
                        cout<<"_   ";
                        ((rtp_hdr_t*)rtp)->m=0;
                    }
                    else
                    {
                        cout<<"i   ";
                        ((rtp_hdr_t*)rtp)->m=1;
                    }
                    switch (slice_type)
                    {
                        case 0: cout<<"P  ";break;
                        case 1: cout<<"B  ";break;
                        case 2: cout<<"I  ";break;
                        case 3: cout<<"SP ";break;
                        case 4: cout<<"SI ";break;
                        case 5: cout<<"P  ";break;
                        case 6: cout<<"B  ";break;
                        case 7: cout<<"I  ";break;
                        case 8: cout<<"SP ";break;
                        case 9: cout<<"SI ";break;
                        default: cout<<"?"<<slice_type<<"? ";
                    }
                    
                    break;
                case 20:
                    cout << "_   SV ";break;
                case 2:
                    cout << "    slice_data_partition_a_layer_rbsp ";break;
                case 3:
                    cout << "    slice_data_partition_b_layer_rbsp ";break;
                case 4:
                    cout << "    slice_data_partition_c_layer_rbsp ";break;
                case 6:
                    cout << "    SEI";break;
                case 7:
                    cout << "    SPS";break;
//                    cout << "    Sequence_parameter_set ";break;
                case 8:
                    cout << "    PPS";break;
//                    cout << "    Picture_parameter_set ";break;
                case 9:
                    cout << "    access_unit_delimiter_rbsp ";break;
                case 10:
                    cout << "    Eend_of_seq_rbsp ";break;
                case 11:
                    cout << "    end_of_stream_rbsp ";break;
                case 12:
                    cout << "    filler_data_rbsp ";break;

                    /*            14     -       Prefix NAL unit                    1.1
                     15     -       Subset sequence parameter set      1.1
                     20     -       Coded slice in scalable extension  1.1
                     30     -       PACSI NAL unit                     4.9
                     31     0       reserved                           4.2.1
                     31     1       Empty NAL unit                     4.10
                     31     2       NI-MTAP                            4.7.1
                     31     3-31    reserved                           4.2.1*/
                case 14:
                    cout << "    svc: Prefix NAL unit ";break;
                case 15:
                    cout << "    svc: Subset sequence parameter set ";break;
                case 30:
                    cout << "    svc: PACSI NAL unit ";break;
                case 31:
                    cout << "    svc: see subtype ";break;
                default:
                    cout << "    Unknown ("<<(nalp[0] & 0x1F)<<")";
            }
            cout << setw(10) << nals;
            

            
/*            0      -          reserved
            1-23   -          Single NAL Unit Packet
            24-27  -          Aggregation Packet
            28-29  -          Fragmentation Unit
            30     -          Single NAL Unit Packet
            31     0          reserved
            31     1          Single NAL Unit Packet
            31     2          Aggregation Packet
            31     3-31       reserved*/
            
        }
        
        cout << endl;
        
    }
    
    
    in_stream.close();
    free(rtp);

    
    return 0;
}

